from . import PropertiesDataBounds


class Coordinate(PropertiesDataBounds):
    '''Mixin class for dimension or auxiliary coordinate constructs.

.. versionadded:: 1.7.0

    '''
    
#--- End: class
